module dataModule
    implicit none
    private 
    public :: putData, data
    type :: data
        integer :: number
    end type data
 contains
    subroutine putData(myData)
        type(data), intent(in) :: myData
        write(*,"(I6)",advance="no") myData%number
    end subroutine
end module dataModule

module doubleLinkedList
    use dataModule
    implicit none
    private 
    public :: list, push, pop, inject, eject, put, deleteAll, getItem, traverse 
    
    type listElem
        type(data) :: content
        type(listElem), pointer :: prev => null(), next => null()
    end type listElem
    type list 
        type(listElem), pointer :: head => null(), tail => null() 
        integer :: length = 0
    end type list
    
 contains

    logical function isEmpty(L)
        type(list), intent(in) :: L
        isEmpty = .not. associated(L%head)
    end function

    integer function length(L)
        type(list), intent(in) :: L 
        length = L%length
    end function

    subroutine push(L, newData)
        type(List), intent(inout) :: L 
        type(data), intent(in) :: newData
        type(listElem), pointer :: tmp
        allocate(tmp)
        tmp%content = newData
        
        if ( isEmpty(L) ) then
            L%tail => tmp
        else 
            tmp%next => L%head
            L%head%prev => tmp
        end if
        
        L%head => tmp
        L%length = L%length + 1
    end subroutine

    subroutine inject(L, newData)
        type(List), intent(inout) :: L 
        type(data), intent(in) :: newData
        type(listElem), pointer :: tmp
        allocate(tmp)
        tmp%content = newData
        
        if ( isEmpty(L) ) then
            L%head => tmp
        else 
            tmp%prev => L%tail
            L%tail%next => tmp
        end if
        
        L%tail => tmp
        L%length = L%length + 1
    end subroutine

    subroutine pop(L, saveData)
        type(list), intent(inout) :: L
        type(data), intent(out), optional :: saveData
        type(listElem), pointer :: tmp
        if ( isEmpty(L) ) then
            write(*,*) "Warning: Cannot push from empty List"
        else
            if ( present(saveData) ) then
                saveData = L%head%content
            end if
            
            if ( length(L) == 1 ) then
                deallocate(L%head)
                L%tail => null()
            else 
                tmp => L%head 
                L%head => L%head%next 
                L%head%prev => null()
                deallocate(tmp)
            end if
            L%length = L%length - 1
        end if
    end subroutine

    subroutine eject(L, saveData)
        type(list), intent(inout) :: L
        type(data), intent(out), optional :: saveData
        type(listElem), pointer :: tmp
        if ( isEmpty(L) ) then
            write(*,*) "Warning: Cannot eject from empty List"
        else
            if ( present(saveData) ) then
                saveData = L%tail%content
            end if
            
            if ( length(L) == 1 ) then
                deallocate(L%head)
                L%tail => null()
            else 
                tmp => L%tail 
                L%tail => L%tail%prev 
                L%tail%next => null()
                deallocate(tmp)
            end if
            L%length = L%length - 1
        end if
    end subroutine

    subroutine put(L)
        type(list), intent(in) :: L
        type(listElem), pointer :: tmp
        integer :: i
        tmp => L%head 
        write(*,"(A)", advance="no") "["
        do i = 1, L%length - 1
            call putData(tmp%content)
            write(*,"(A)", advance="no") ", "
            tmp => tmp%next 
        end do 
        if (associated(tmp)) then 
            call putData(tmp%content)
        end if
        write(*,"(A)") "]"
    end subroutine

    subroutine deleteAll(L)
        type(list), intent(inout) :: L
        do while(.not. isEmpty(L))
            call pop(L)
        end do 
    end subroutine

    subroutine getItem(L, i, saveData)
        type(list), intent(in) :: L
        type(listElem), pointer :: tmp
        type(data), intent(out) :: saveData
        integer, intent(in) :: i
        integer :: k 
        if ( i > 0 .and. i<=L%length ) then
            tmp => L%head
            do k=1,i-1
                tmp => tmp%next
            end do 
            saveData = tmp%content
        else 
            write(*,*) "Index out of bounds in getItem."
        end if
    end subroutine

    subroutine traverse(L, task)
        type(list), intent(in) :: L
        type(listElem), pointer :: tmp
        integer :: i
        interface 
            subroutine task(elem)
                use dataModule
                type(data), intent(in) :: elem
            end subroutine
        end interface 
        tmp => L%head
        do i = 1, L%length
            call task(tmp%content)
            tmp => tmp%next
        end do 
    end subroutine
end module doubleLinkedList

module highLevelListFunctions
    use dataModule
    use doubleLinkedList
    implicit none
    private 
    public :: map, foldLeft, filter
contains
    subroutine mapSubroutine(inputList, someFunction, outputList)
        type(list), intent(in) :: inputList
        type(list), intent(out) :: outputList
        interface
            function someFunction(someData)
                use dataModule
                type(data), intent(in) :: someData
                type(data) :: someFunction
            end function someFunction
        end interface
        call traverse(inputList,taskForMap)
    contains
        subroutine taskForMap(someData)
            type(data), intent(in) :: someData
            call inject(outputList, someFunction(someData))
        end subroutine taskForMap
    end subroutine mapSubroutine

    function map(inputList,someFunction) result(outputList)
        type(list), intent(in) :: inputList
        type(list) :: outputList
        interface
            function someFunction(someData)
                use dataModule
                type(data), intent(in) :: someData
                type(data) :: someFunction
            end function someFunction
        end interface
        call mapSubroutine(inputList,someFunction,outputList)
    end function map

    subroutine foldLeftSubroutine(someList, someFunction, someStartValue, outputValue) 
        type(list), intent(in) :: someList
        type(data), intent(in) :: someStartValue
        type(data), intent(out) :: outputValue
        interface
            function someFunction(someData, otherData)
                use dataModule
                type(data), intent(in) :: someData, otherData
                type(data) :: someFunction
            end function someFunction
        end interface
        outputValue = someStartValue
        call traverse(someList, taskFoldLeft)
    contains
        subroutine taskFoldLeft (someData)
            type(data), intent(in) :: someData
            outputValue = someFunction(outputValue,someData)
        end subroutine taskFoldLeft
    end subroutine foldLeftSubroutine

    function foldLeft(someList,someFunction,someStartValue) result(outputValue)
        type(list), intent(in) :: someList
        type(data), intent(in) :: someStartValue
        type(data) :: outputValue
        interface
            function someFunction(someData, otherData)
                use dataModule
                type(data), intent(in) :: someData, otherData
                type(data) :: someFunction
            end function someFunction
        end interface
        call foldLeftSubroutine(someList,someFunction,someStartValue,outputValue)
    end function foldLeft

    subroutine filterSubroutine(inputList,someCondition,outputList)
        type(list), intent(in) :: inputList
        type(list), intent(out) :: outputList
        interface
            logical function someCondition(someData)
                use dataModule
                type(data), intent(in) :: someData
            end function someCondition
        end interface
        call traverse(inputList,taskFilter)
    contains 
        subroutine taskFilter(someData)
            type(data), intent(in) :: someData
            if ( someCondition(someData) ) then
                call inject(outputList, someData)
            end if
        end subroutine taskFilter
    end subroutine filterSubroutine

    function filter(inputList, someCondition) result(outputList)
        type(list), intent(in) :: inputList
        type(list) :: outputList
        interface
            logical function someCondition(someData)
                use dataModule
                type(data), intent(in) :: someData
            end function someCondition
        end interface
        call filterSubroutine(inputList, someCondition, outputList)
    end function filter
end module highLevelListFunctions

program main
    use dataModule
    use doubleLinkedList
    use highLevelListFunctions
    implicit none
    type(list) :: L
    type(data) :: myData
    integer :: i
    
    ! Building a list
    do i=1,5
        myData%number = i 
        call push(L, myData)
        myData%number = 10*i
        call inject(L, myData)
    end do
    call put(L)
    
    call put(map(L,square))

    myData%number = 0
    write(*,*) foldleft(L,add,myData)

    call put(filter(L,even))
    
    call deleteAll(L)
    
 contains 
    subroutine collect(elem, L2)
        type(data), intent(in) :: elem
        type(list), intent(inout) :: L2
        call push(L2, elem)
    end subroutine

    function square(someData)
        type(data), intent(in) :: someData
        type(data) :: square
        square%number = (someData%number)**2
    end function square

    function add(a, b)
        type(data), intent(in) :: a, b 
        type(data) :: add 
        add%number = a%number + b%number
    end function

    logical function even(someData) 
        type(Data),intent(in)::someData
        even= mod(someData%number,2)==0
    end function even
    
end program main